package de.juplo.demo;


import java.net.URI;
import java.nio.charset.Charset;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.IntPredicate;
import java.util.function.Predicate;
import java.util.function.Supplier;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.when;
import org.reactivestreams.Publisher;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.http.client.reactive.ClientHttpRequest;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.util.MultiValueMap;
import org.springframework.web.reactive.function.BodyInserter;
import org.springframework.web.reactive.function.client.ClientResponse;
import org.springframework.web.reactive.function.client.WebClient;
import org.springframework.web.reactive.function.client.WebClient.RequestHeadersSpec;
import org.springframework.web.reactive.function.client.WebClient.RequestHeadersUriSpec;
import org.springframework.web.reactive.function.client.WebClient.ResponseSpec;
import org.springframework.web.util.UriBuilder;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;


/**
 * Unit-Test for class {@link RemoteContentService}.
 * @author Kai Moritz
 */
@ExtendWith(SpringExtension.class)
public class RemoteContentServiceTest
{
  RemoteContentService service;

  @MockBean
  WebClient webClient;
  @MockBean(name = "uriSpec")
  RequestHeadersUriSpec uriSpec;
  @MockBean(name = "headersSpec")
  RequestHeadersSpec headersSpec;
  @MockBean(name = "responseSpec")
  ResponseSpec responseSpec;


  @BeforeEach
  void setUp()
  {
    service = new RemoteContentService(webClient);
  }


  @Test
  void test()
  {
    when(webClient.get()).thenReturn(uriSpec);
    when(uriSpec.uri(eq("/foo"))).thenReturn(headersSpec);
    when(headersSpec.retrieve()).thenReturn(responseSpec);
    when(responseSpec.bodyToMono(String.class)).thenReturn(Mono.just("bar"));

    Mono<String> result = service.getRemoteText("/foo");

    StepVerifier
        .create(result)
        .expectNext("bar")
        .verifyComplete();
  }


  static class WebClientStub implements WebClient
  {
    final static WebClientStub WEB_CLIENT = new WebClientStub();
    final static RequestHeadersUriSpecStub HEADERS_SPEC = new RequestHeadersUriSpecStub();
    final static RequestBodySpecStub REQUEST_BODY_SPEC = new RequestBodySpecStub();
    final static ResponseSpecStub RESPONSE_SPEC = new ResponseSpecStub();

    static Supplier<Mono<?>> toMono;
    static Supplier<Flux<?>> toFlux;


    private WebClientStub() {}


    @Override
    public RequestHeadersUriSpec<?> get()
    {
      return HEADERS_SPEC;
    }

    @Override
    public RequestHeadersUriSpec<?> head()
    {
      return HEADERS_SPEC;
    }

    @Override
    public RequestBodyUriSpec post()
    {
      return REQUEST_BODY_SPEC;
    }

    @Override
    public RequestBodyUriSpec put()
    {
      return REQUEST_BODY_SPEC;
    }

    @Override
    public RequestBodyUriSpec patch()
    {
      return REQUEST_BODY_SPEC;
    }

    @Override
    public RequestHeadersUriSpec<?> delete()
    {
      return HEADERS_SPEC;
    }

    @Override
    public RequestHeadersUriSpec<?> options()
    {
      return HEADERS_SPEC;
    }

    @Override
    public RequestBodyUriSpec method(HttpMethod hm)
    {
      return REQUEST_BODY_SPEC;
    }

    @Override
    public Builder mutate()
    {
      throw new UnsupportedOperationException("Stub is immutable!");
    }


    static class RequestHeadersUriSpecStub implements RequestHeadersUriSpec
    {
      private RequestHeadersUriSpecStub() {}


      @Override
      public RequestHeadersSpec uri(URI uri)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec uri(String string, Object... os)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec uri(String string, Map map)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec uri(String string, Function fnctn)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec uri(Function fnctn)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec accept(MediaType... mts)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec acceptCharset(Charset... chrsts)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec cookie(String string, String string1)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec cookies(Consumer cnsmr)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec ifModifiedSince(ZonedDateTime zdt)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec ifNoneMatch(String... strings)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec header(String string, String... strings)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec headers(Consumer cnsmr)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec attribute(String string, Object o)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec attributes(Consumer cnsmr)
      {
        return this;
      }

      @Override
      public ResponseSpec retrieve()
      {
        return RESPONSE_SPEC;
      }

      @Override
      public Mono exchange()
      {
        return WebClientStub.toMono.get();
      }
    }

    static class RequestBodySpecStub implements RequestBodyUriSpec
    {
      private RequestBodySpecStub() {}


      @Override
      public RequestBodySpec contentLength(long l)
      {
        return this;
      }

      @Override
      public RequestBodySpec contentType(MediaType mt)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec<?> bodyValue(Object o)
      {
        return this;
      }

      @Override
      public <T, P extends Publisher<T>> RequestHeadersSpec<?> body(P p, Class<T> type)
      {
        return this;
      }

      @Override
      public <T, P extends Publisher<T>> RequestHeadersSpec<?> body(P p, ParameterizedTypeReference<T> ptr)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec<?> body(Object o, Class<?> type)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec<?> body(Object o, ParameterizedTypeReference<?> ptr)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec<?> body( BodyInserter<?, ? super ClientHttpRequest> bi)
      {
        return this;
      }

      @Override
      public RequestHeadersSpec<?> syncBody(Object o)
      {
        return this;
      }

      @Override
      public RequestBodySpec accept(MediaType... mts)
      {
        return this;
      }

      @Override
      public RequestBodySpec acceptCharset(Charset... chrsts)
      {
        return this;
      }

      @Override
      public RequestBodySpec cookie(String string, String string1)
      {
        return this;
      }

      @Override
      public RequestBodySpec cookies(Consumer<MultiValueMap<String, String>> cnsmr)
      {
        return this;
      }

      @Override
      public RequestBodySpec ifModifiedSince(ZonedDateTime zdt)
      {
        return this;
      }

      @Override
      public RequestBodySpec ifNoneMatch(String... strings)
      {
        return this;
      }

      @Override
      public RequestBodySpec header(String string, String... strings)
      {
        return this;
      }

      @Override
      public RequestBodySpec headers(Consumer<HttpHeaders> cnsmr)
      {
        return this;
      }

      @Override
      public RequestBodySpec attribute(String string, Object o)
      {
        return this;
      }

      @Override
      public RequestBodySpec attributes(Consumer<Map<String, Object>> cnsmr)
      {
        return this;
      }

      @Override
      public ResponseSpec retrieve()
      {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
      }

      @Override
      public Mono<ClientResponse> exchange()
      {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
      }

      @Override
      public RequestBodySpec uri(URI uri)
      {
        return this;
      }

      @Override
      public RequestBodySpec uri(String string, Object... os)
      {
        return this;
      }

      @Override
      public RequestBodySpec uri(String string, Map<String, ?> map)
      {
        return this;
      }

      @Override
      public RequestBodySpec uri(String string, Function<UriBuilder, URI> fnctn)
      {
        return this;
      }

      @Override
      public RequestBodySpec uri(Function<UriBuilder, URI> fnctn)
      {
        return this;
      }
    }

    static class ResponseSpecStub implements ResponseSpec
    {
      private ResponseSpecStub() {}


      @Override
      public ResponseSpec onStatus(Predicate<HttpStatus> statusPredicate, Function<ClientResponse, Mono<? extends Throwable>> exceptionFunction)
      {
        return this;
      }

      @Override
      public ResponseSpec onRawStatus(IntPredicate statusCodePredicate, Function<ClientResponse, Mono<? extends Throwable>> exceptionFunction)
      {
        return this;
      }

      @Override
      public <T> Mono<T> bodyToMono(Class<T> elementClass)
      {
        return (Mono<T>)toMono.get();
      }

      @Override
      public <T> Mono<T> bodyToMono(ParameterizedTypeReference<T> elementTypeRef)
      {
        return (Mono<T>)toMono.get();
      }

      @Override
      public <T> Flux<T> bodyToFlux(Class<T> elementClass)
      {
        return (Flux<T>)toFlux;
      }

      @Override
      public <T> Flux<T> bodyToFlux(ParameterizedTypeReference<T> elementTypeRef)
      {
        return (Flux<T>)toFlux;
      }

      @Override
      public <T> Mono<ResponseEntity<T>> toEntity(Class<T> bodyClass)
      {
        return (Mono<ResponseEntity<T>>)toMono.get();
      }

      @Override
      public <T> Mono<ResponseEntity<T>> toEntity(ParameterizedTypeReference<T> bodyTypeReference)
      {
        return (Mono<ResponseEntity<T>>)toMono.get();
      }

      @Override
      public <T> Mono<ResponseEntity<List<T>>> toEntityList(Class<T> elementClass)
      {
        return (Mono<ResponseEntity<List<T>>>)toMono.get();
      }

      @Override
      public <T> Mono<ResponseEntity<List<T>>> toEntityList(ParameterizedTypeReference<T> elementTypeRef)
      {
        return (Mono<ResponseEntity<List<T>>>)toMono.get();
      }

      @Override
      public Mono<ResponseEntity<Void>> toBodilessEntity()
      {
        return (Mono<ResponseEntity<Void>>)toMono.get();
      }
    }
  }
}
